//
//   Snapping.cpp
//
//
#include <math.h>
#include "wx/wx.h"
#include "KidsCindyCanvas.h"
#include "resource.h"
#include "Meet.h"

double abs(double a);

bool MyCanvas::Snapping(GPoint* p, double* gx,double* gy,GPoint** pt)
{
	if(p==NULL){
		*pt=NULL;
		return false;
	}
	double geomx=p->xz().real();
	double geomy=p->yz().real();
	if((geomx - *gx)*(geomx - *gx) + (geomy - *gy)*(geomy - *gy) < snapradius2 && p->invisible==false){
		*gx=geomx;
		*gy=geomy;
		*pt=p;
		return true;
	}
	if(p->next!=NULL){
		return Snapping(p->next,gx,gy,pt);
	}
	*pt=NULL;
	return false;
}

bool MyCanvas::NearPoint(GPoint* p, double* gx,double* gy,GPoint** pt)
{
	if(p==NULL){
		*pt=NULL;
		return false;
	}
	double geomx=p->xz().real();
	double geomy=p->yz().real();
	if((geomx - *gx)*(geomx - *gx) + (geomy - *gy)*(geomy - *gy) < snapradius2 && p->invisible==false){
		*pt=p;
		return true;
	}
	if(p->next!=NULL){
		return NearPoint(p->next,gx,gy,pt);
	}
	*pt=NULL;
	return false;
}

bool MyCanvas::Snap2Line(GLine* lines, double* gx, double* gy)
{
	if(lines==NULL){
		return false;
	}
	double a,b,c;
	if(lines->realp(&a,&b,&c)==false){
		return false;
	}
	double x0=-(a*c-b*(b*(*gx)-a*(*gy)))/(a*a+b*b);
	double y0=-(a*(b*(*gx)-a*(*gy))+b*c)/(a*a+b*b);
	if((x0- *gx)*(x0 - *gx)+(y0 - *gy)*(y0 - *gy) < snapradius2  && lines->invisible==false){
		*gx=x0;
		*gy=y0;
		return true;
	}
	if(lines->next!=NULL){
		return Snap2Line(lines->next,gx,gy);
	}
	return false;
}

int MyCanvas::Snap2LineEx(GLine* lines, double* gx,double* gy, GLine** ln1, GLine** ln2)
{
	if(lines==NULL){
		return 0;
	}
	int lncount=0;
	double a,b,c;
	if(lines->realp(&a,&b,&c)==false){
		return 0;
	}
	if(lines->next!=NULL){
		lncount = Snap2LineEx(lines->next,gx,gy,ln1,ln2);
	}
	double x0=-(a*c-b*(b*(*gx)-a*(*gy)))/(a*a+b*b);
	double y0=-(a*(b*(*gx)-a*(*gy))+b*c)/(a*a+b*b);
	if((x0- *gx)*(x0 - *gx)+(y0 - *gy)*(y0 - *gy) < snapradius2 && lines->invisible==false){
		if(lncount==0){
			*gx=x0;
			*gy=y0;
			*ln1=lines;
			*ln2=NULL;
			return 1;
		}
		else if(lncount==1){
			GPoint gp=Meet(lines,*ln1);
			*gx = gp.xz().real();
			*gy = gp.yz().real();
			*ln2=lines;
			return 2;
		}
	}
	return lncount;
}

bool MyCanvas::Snap2Circle(GCircle* circles, double* gx, double* gy)
{
	if(circles==NULL){
		return false;
	}
	double a,b,c;
	if(circles->realp(&a,&b,&c)==false){
		return false;
	}
	double r=sqrt((a - *gx)*(a - *gx)+(b - *gy)*(b - *gy));
	if( sqrt(c) - snapradius < r && r < sqrt(c) + snapradius && circles!=cursorci  && circles->invisible==false){
		Proj2Circle(circles,gx,gy);
		return true;
	}
	if(circles->next!=NULL){
		return Snap2Circle(circles->next,gx,gy);
	}
	return false;
}

int MyCanvas::Snap2CircleEx(GCircle* circles, double* gx,double* gy, GCircle** ci1, GCircle** ci2)
{
	if(circles==NULL){
		return 0;
	}
	int cicount=0;
	double a,b,c;
	if(circles->realp(&a,&b,&c)==false){
		return 0;
	}
	if(circles->next!=NULL){
		cicount = Snap2CircleEx(circles->next,gx,gy,ci1,ci2);
	}
	double r=sqrt((a - *gx)*(a - *gx)+(b - *gy)*(b - *gy));
	if( sqrt(c) - snapradius < r && r < sqrt(c) + snapradius && circles!=cursorci  && circles->invisible==false){
		if(cicount==0){
			Proj2Circle(circles,gx,gy);
			*ci1=circles;
			*ci2=NULL;
			return 1;
		}
		else if(cicount==1){
			Meet(circles,*ci1,gx,gy);
			*ci2=circles;
			return 2;
		}
	}
	return cicount;
}

void MyCanvas::PtDragHilight(GLine* ln, double gx,double gy)
{
	if(ln==NULL){
		return;
	}
	double a,b,c;
	if(ln->realp(&a,&b,&c)==false){
		ln->hilighted=false;
		return;
	}
	double x0=-(a*c-b*(b*gx-a*gy))/(a*a+b*b);
	double y0=-(a*(b*gx-a*gy)+b*c)/(a*a+b*b);
	if((x0- gx)*(x0 - gx)+(y0 - gy)*(y0 - gy) < snapradius2){
		ln->hilighted=true;
	}
	else if(!ShiftLButton) {
		ln->hilighted=false;
	}
	PtDragHilight(ln->next,gx,gy);
}

void MyCanvas::PtDragHilight(GCircle* ci, double gx,double gy)
{
	if(ci==NULL){
		return;
	}
	double a,b,c;
	if(ci->realp(&a,&b,&c)==false){
		ci->hilighted=false;
		return;
	}
	double r=sqrt((a - gx)*(a - gx)+(b - gy)*(b - gy));
	if( sqrt(c) - snapradius < r && r < sqrt(c) + snapradius){
		ci->hilighted=true;
	}
	else if(!ShiftLButton) {
		ci->hilighted=false;
	}
	PtDragHilight(ci->next,gx,gy);
}

void MyCanvas::PtDragHilight(GPoint* pt, double gx,double gy)
{
	if(pt==NULL){
		return;
	}
	double geomx=pt->xz().real();
	double geomy=pt->yz().real();
	if((geomx - gx)*(geomx - gx) + (geomy - gy)*(geomy - gy) < snapradius2){
		pt->hilighted=true;
	}
	else if(!ShiftLButton){
		pt->hilighted=false;
	}
	PtDragHilight(pt->next,gx,gy);
}

void MyCanvas::PtDragHilight(GParaCurve* cu, double gx,double gy)
{
	if(cu==NULL){
		return;
	}
	if(cu->orbit_len<=1){
		return;
	}
	double g1x,g1y,g2x,g2y;
	double x0,y0,t=0.0;
	bool snap=false;
	int i;
	g1x=cu->orbitx[0];
	g1y=cu->orbity[0];
	for(i=1;i<cu->orbit_len && !snap;i++){
		g2x=cu->orbitx[i];
		g2y=cu->orbity[i];
		if(g1x!=g2x || g1y!=g2y){
			t = ((gx-g1x)*(g2x-g1x)+(gy-g1y)*(g2y-g1y)) / ((g2x-g1x)*(g2x-g1x)+(g2y-g1y)*(g2y-g1y));
			x0 = g1x*(1-t)+g2x*t;
			y0 = g1y*(1-t)+g2y*t;
			if(t>=0.0 && t<1.0 && (x0- gx)*(x0 - gx)+(y0 - gy)*(y0 - gy) < snapradius2  && cu->invisible==false){
				snap=true;
			}
		}
		g1x=g2x;g1y=g2y;
	}
	if(snap){
		cu->hilighted=true;
	}
	else if(!ShiftLButton){
		cu->hilighted=false;
	}
	PtDragHilight(cu->next,gx,gy);
}

void MyCanvas::SelectHilight(GPoint* pt, double gx,double gy)
{
	if(pt==NULL){
		return;
	}
	double geomx=pt->xz().real();
	double geomy=pt->yz().real();
	if((geomx - gx)*(geomx - gx) + (geomy - gy)*(geomy - gy) < snapradius2 && pt->invisible==false){
		if(!ShiftLButton){
			ClearAllHilights();
		}
		Sethilighted(pt);
	}
	else if(!ShiftLButton){
		Deletehilighted(pt);
	}
	SelectHilight(pt->next,gx,gy);
}

void MyCanvas::SelectHilight(GLine* ln, double gx,double gy)
{
	if(ln==NULL){
		return;
	}
	double a,b,c;
	if(ln->realp(&a,&b,&c)==false){
		ln->hilighted=false;
		return;
	}
	double x0=-(a*c-b*(b*gx-a*gy))/(a*a+b*b);
	double y0=-(a*(b*gx-a*gy)+b*c)/(a*a+b*b);
	if((x0- gx)*(x0 - gx)+(y0 - gy)*(y0 - gy) < snapradius2){
		if(!ShiftLButton){
			ClearAllHilights();
		}
		Sethilighted(ln);
		if(GMode==MODE_CINDY){
			ln->clipping=LineClipping;
		}
	}
	else if(!ShiftLButton){
		Deletehilighted(ln);
	}
	SelectHilight(ln->next,gx,gy);
}

void MyCanvas::SelectHilight(GCircle* ci, double gx,double gy)
{
	if(ci==NULL){
		return;
	}
	double a,b,c;
	if(ci->realp(&a,&b,&c)==false){
		ci->hilighted=false;
		return;
	}
	double r=sqrt((a - gx)*(a - gx)+(b - gy)*(b - gy));
	if( sqrt(c) - snapradius < r && r < sqrt(c) + snapradius){
		if(!ShiftLButton){
			ClearAllHilights();
		}
		Sethilighted(ci);
	}
	else if(!ShiftLButton){
		Deletehilighted(ci);
	}
	SelectHilight(ci->next,gx,gy);
}

bool MyCanvas::Snap2Text(GText* text, int px,int py)
{
	if(text==NULL){
		return false;
	}
	int length,height;
	wxPaintDC dc(this);
	PrepareDC(dc);
	wxFont* myfont = wxFont::New((int)floor(text->fontsize), wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false, wxT(""),  wxFONTENCODING_DEFAULT);
	dc.SetFont(*myfont);
	GetMultiTextSize((wxString)text->str,(wxMemoryDC*)&dc,&length,&height);
	delete myfont;
	int tx=(int)floor(text->tx);
	int ty=(int)floor(text->ty);
	if(tx<px && px<tx+length && ty<py && py<ty+height  && text->invisible==false){
		cursortext=text;
		cursortext->m_text_dragging_h_length=px - tx;
		cursortext->m_text_dragging_h_height=py - ty;
		return true;
	}
	else {
		cursortext=NULL;
		return Snap2Text(text->next,px,py);
	}
}

int MyCanvas::Snap2Image(GImage* im, int px, int py)
{
	if(im==NULL){
		return 0;
	}
	//printf("px=%d,py=%d,m_x=%d,m_y=%d\n",px,py,im->m_x,im->m_y);
	if(im->m_x <= px && px <= im->m_x+im->m_w && im->m_y <= py && py <= im->m_y+im->m_h && !(im->invisible)){
		cursorim=im;
		im->m_image_drag_x = px - im->m_x;
		im->m_image_drag_y = py - im->m_y;
		return 1;
	}
	else if(im->m_x-5 <= px && px <= im->m_x && im->m_y-5 <= py && py <= im->m_y && !(im->invisible)){
		cursorim=im;
		im->m_image_drag_x = px - im->m_x;
		im->m_image_drag_y = py - im->m_y;
		return 2;
	}
	else if(im->m_x-5 <= px && px <= im->m_x && im->m_y+im->m_h <= py && py <= im->m_y+im->m_h+5 && !(im->invisible)){
		cursorim=im;
		im->m_image_drag_x = px - im->m_x;
		im->m_image_drag_y = py - im->m_y - im->m_h;
		return 3;
	}
	else if(im->m_x+im->m_w <= px && px <= im->m_x+im->m_w+5 && im->m_y-5 <= py && py <= im->m_y && !(im->invisible)){
		cursorim=im;
		im->m_image_drag_x = px - im->m_x - im->m_w;
		im->m_image_drag_y = py - im->m_y;
		return 4;
	}
	else if(im->m_x+im->m_w <= px && px <= im->m_x+im->m_w+5 && im->m_y+im->m_h <= py && py <= im->m_y+im->m_h+5 && !(im->invisible)){
		cursorim=im;
		im->m_image_drag_x = px - im->m_x - im->m_w;
		im->m_image_drag_y = py - im->m_y - im->m_h;
		return 5;
	}
	else {
		cursorim=NULL;
		return Snap2Image(im->next,px,py);
	}
}

bool MyCanvas::Snap2Curve(GParaCurve* curve, double* gx,double *gy)
{
	if(curve==NULL){
		return false;
	}
	int i;
	if(curve->orbit_len<=1){
		return false;
	}
	double g1x,g1y,g2x,g2y;
	double x0,y0,t;
	g1x=curve->orbitx[0];
	g1y=curve->orbity[0];
	for(i=1;i<curve->orbit_len;i++){
		g2x=curve->orbitx[i];
		g2y=curve->orbity[i];
		if(g1x!=g2x || g1y!=g2y){
			t = ((*gx-g1x)*(g2x-g1x)+(*gy-g1y)*(g2y-g1y)) / ((g2x-g1x)*(g2x-g1x)+(g2y-g1y)*(g2y-g1y));
			x0 = g1x*(1-t)+g2x*t;
			y0 = g1y*(1-t)+g2y*t;
			if(t>=0.0 && t<1.0 && (x0- *gx)*(x0 - *gx)+(y0 - *gy)*(y0 - *gy) < snapradius2  && curve->invisible==false){
				*gx=x0;
				*gy=y0;
				return true;
			}
		}
		g1x=g2x;
		g1y=g2y;
	}
	if(curve->next!=NULL){
		return Snap2Curve(curve->next,gx,gy);
	}
	return false;
}

int MyCanvas::Snap2CurveEx(GParaCurve* cu, double* gx,double* gy, GParaCurve** cu1, GParaCurve** cu2)
{
	if(cu==NULL){
		return 0;
	}
	int cucount=0;
	if(cu->next!=NULL){
		cucount = Snap2CurveEx(cu->next,gx,gy,cu1,cu2);
		if(cucount>=2){
			return cucount;
		}
	}
	double g1x,g1y,g2x,g2y;
	double x0=0.0,y0=0.0,t=0.0;
	bool snap=false;
	int i;
	g1x=cu->orbitx[0];
	g1y=cu->orbity[0];
	for(i=1;i<cu->orbit_len && !snap;i++){
		g2x=cu->orbitx[i];
		g2y=cu->orbity[i];
		if(g1x!=g2x || g1y!=g2y){
			t = ((*gx-g1x)*(g2x-g1x)+(*gy-g1y)*(g2y-g1y)) / ((g2x-g1x)*(g2x-g1x)+(g2y-g1y)*(g2y-g1y));
			x0 = g1x*(1-t)+g2x*t;
			y0 = g1y*(1-t)+g2y*t;
			if(t>=0.0 && t<1.0 && (x0- *gx)*(x0 - *gx)+(y0 - *gy)*(y0 - *gy) < snapradius2  && cu->invisible==false){
				snap=true;
			}
		}
		g1x=g2x;g1y=g2y;
	}
	if(snap){
		if(cucount==0){
			*gx=x0;
			*gy=y0;
			*cu1=cu;
			*cu2=NULL;
			return 1;
		}
		else if(cucount==1){
			*gx=x0;
			*gy=y0;
			*cu2=cu;
			return 2;
		}
	}
	return cucount;
}

bool MyCanvas::Snap2PtName(GPoint* pt, int px,int py)
{
	if(pt==NULL){
		return false;
	}
	long length,height,descent;
	wxPaintDC dc(this);
	PrepareDC(dc);
	dc.SetFont(wxFont(pt->_fontsize,wxFONTFAMILY_DEFAULT,wxFONTSTYLE_NORMAL,wxFONTWEIGHT_NORMAL));
	dc.GetTextExtent(pt->_name,&length,&height,&descent);
	int tx = centerx + (int)floor((pt->xz().real()-geomcx)/geomperdot)+(int)floor((10+pt->_fontsize)*cos(pt->_name_theta))-(pt->_fontsize/2);
	int ty = centery - (int)floor((pt->yz().real()-geomcy)/geomperdot)-(int)floor((10+pt->_fontsize)*sin(pt->_name_theta))-(pt->_fontsize);
	if(tx<px && px<tx+length && ty<py && py<ty+height+descent && pt->invisible==false){
		cursorpt=pt;
		return true;
	}
	else {
		cursorpt=NULL;
		return Snap2PtName(pt->next,px,py);
	}
}

void MyCanvas::SelectHilight(GParaCurve* curve, double gx,double gy)
{
	if(curve==NULL){
		return;
	}
	bool good;
	int i;
	double g1x,g1y,g2x,g2y;
	double x0,y0,t;
	if(curve->orbit_len<=1){
		good=false;
	}
	else {
		good=false;
		g1x=curve->orbitx[0];
		g1y=curve->orbity[0];
		for(i=1;i<curve->orbit_len && good==false;i++){
			g2x=curve->orbitx[i];
			g2y=curve->orbity[i];
			if(g1x!=g2x || g1y!=g2y){
				t = ((gx-g1x)*(g2x-g1x)+(gy-g1y)*(g2y-g1y)) / ((g2x-g1x)*(g2x-g1x)+(g2y-g1y)*(g2y-g1y));
				x0 = g1x*(1-t)+g2x*t;
				y0 = g1y*(1-t)+g2y*t;
				if(t>=0.0 && t<1.0 && (x0- gx)*(x0 - gx)+(y0 - gy)*(y0 - gy) < snapradius2){
					good=true;
				}
			}
			g1x=g2x;
			g1y=g2y;
		}
	}
	if(good){
		if(!ShiftLButton){
			ClearAllHilights();
		}
		Sethilighted(curve);
	}
	else if(!ShiftLButton){
		Deletehilighted(curve);
	}
	SelectHilight(curve->next,gx,gy);
}

//bool MyCanvas::Snap2Angle(GAngle* ag, double* gx, double* gy)
//{
//	if(ag==NULL){
//		*gx=*gy=0.0;
//		return false;
//	}
//	double ax = *gx - ag->m_xx.real();
//	double ay = *gy - ag->m_yy.real();
//	double inners = ax * ag->m_startx.real() + ay * ag->m_starty.real(), innere = ax * ag->m_endx.real() + ay * ag->m_endy.real();
//	double outers = ax * ag->m_starty.real() - ay * ag->m_startx.real(), outere = ax * ag->m_endy.real() - ay * ag->m_endx.real();
//	double norma = sqrt(ax * ax + ay * ay) / geomperdot / ag->m_angle_size;
//	if(inners >0.01 && innere>0.01 && outers * outere < -0.01 && 0.8 < norma && norma < 1.2   && ag->invisible==false){
//		return true;
//	}
//	if(ag->next!=NULL){
//		return Snap2Angle(ag->next,gx,gy);
//	}
//	return false;
//}

bool MyCanvas::SelectHilight(GAngle* ag, double gx,double gy)
{
	if(ag==NULL){
		cursorag=NULL;
		return false;
	}
	double ax = gx - ag->m_xx.real();
	double ay = gy - ag->m_yy.real();
	double inners = ax * ag->m_startx.real() + ay * ag->m_starty.real(), innere = ax * ag->m_endx.real() + ay * ag->m_endy.real();
	double outers = ax * ag->m_starty.real() - ay * ag->m_startx.real(), outere = ax * ag->m_endy.real() - ay * ag->m_endx.real();
	double norma = sqrt(ax * ax + ay * ay) / geomperdot / ag->m_angle_size;
	if(inners >0.01 && innere>0.01 && outers * outere < -0.01 && 0.8 < norma && norma < 1.2   && ag->invisible==false){
		if(!ShiftLButton){
			ClearAllHilights();
		}
		Sethilighted(ag);
		cursorag=ag;
		return true;
	}
	return SelectHilight(ag->next,gx,gy);

}

bool MyCanvas::Snap2Bracket(GBracket* br,double gx,double gy)
{
	if(br==NULL){
		cursorbr=NULL;
		return false; 
	}
	CComplex sxz=br->m_startx/br->m_startz;
	CComplex syz=br->m_starty/br->m_startz;
	CComplex exz=br->m_endx/br->m_endz;
	CComplex eyz=br->m_endy/br->m_endz;
	if(!(sxz.realp()) || !(syz.realp()) || !(exz.realp()) || !(eyz.realp())){
		return Snap2Bracket(br->next,gx,gy);
	}
	double s_geomx=sxz.real();
	double s_geomy=syz.real();
	double e_geomx=exz.real();
	double e_geomy=eyz.real();
	double d=sqrt((s_geomx-e_geomx)*(s_geomx-e_geomx)+(s_geomy-e_geomy)*(s_geomy-e_geomy));
	if(d==0.0){
		return  Snap2Bracket(br->next,gx,gy);
	}
	double sn=(br->m_bracket_height>0)? 1.0 : -1.0;
	double dh=abs(d*br->m_bracket_height);
	double ax=0.5*(s_geomx+e_geomx);
	double ay=0.5*(s_geomy+e_geomy);
	double cx=ax-sn*(s_geomy-e_geomy)/d*dh;
	double cy=ay+sn*(s_geomx-e_geomx)/d*dh;
	if(gx>cx-snapradius && gx<cx+snapradius && gy>cy-snapradius && gy<cy+snapradius){
		cursorbr=br;
		return true;
	}
	return Snap2Bracket(br->next,gx,gy);
}



